# Generated by Django 3.2.15 on 2022-09-29 17:20

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('integrated_channel', '0020_auto_20220929_1712'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='contentmetadataitemtransmission',
            name='api_response_body',
        ),
    ]



# Generated by Django 3.2.20 on 2023-08-22 11:48

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0179_restore_historical_table'),
    ]

    operations = [
        migrations.CreateModel(
            name='ChatGPTResponse',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('prompt', models.TextField(help_text='ChatGPT prompt.')),
                ('prompt_hash', models.CharField(editable=False, max_length=32)),
                ('response', models.TextField(help_text='ChatGPT response.')),
                ('enterprise_customer', models.ForeignKey(help_text='The enterprise that can be linked using this key.', on_delete=django.db.models.deletion.CASCADE, related_name='chat_gpt_prompts', to='enterprise.enterprisecustomer')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



"""
Send xAPI statements to the LRS configured via admin.
"""

from logging import getLogger

from django.contrib import auth
from django.core.management.base import BaseCommand, CommandError

from enterprise.api_client.discovery import get_course_catalog_api_service_client
from enterprise.models import EnterpriseCourseEnrollment, EnterpriseCustomer, EnterpriseCustomerUser
from enterprise.utils import NotConnectedToOpenEdX
from integrated_channels.xapi.models import XAPILearnerDataTransmissionAudit, XAPILRSConfiguration
from integrated_channels.xapi.utils import is_success_response, send_course_completion_statement

try:
    from lms.djangoapps.grades.models import PersistentCourseGrade
except ImportError:
    PersistentCourseGrade = None

try:
    from openedx.core.djangoapps.content.course_overviews.models import CourseOverview
except ImportError:
    CourseOverview = None

LOGGER = getLogger(__name__)
User = auth.get_user_model()


class Command(BaseCommand):
    """
    Send course completion xAPI statements to enterprise customers.
    """

    def add_arguments(self, parser):
        """
        Add required arguments to the parser.
        """
        parser.add_argument(
            '--days',
            dest='days',
            required=False,
            type=int,
            default=1,
            help='Send xAPI analytics for learners who enrolled during last this number of days.'
        )
        parser.add_argument(
            '--enterprise_customer_uuid',
            dest='enterprise_customer_uuid',
            type=str,
            required=False,
            help='Send xAPI analytics for this enterprise customer only.'
        )
        super().add_arguments(parser)

    @staticmethod
    def parse_arguments(*args, **options):
        """
        Parse and validate arguments for the command.

        Arguments:
            *args: Positional arguments passed to the command
            **options: Optional arguments passed to the command

        Returns:
            A tuple containing parsed values for
            1. days (int): Integer showing number of days to lookup enterprise enrollments,
                course completion etc and send to xAPI LRS
            2. enterprise_customer_uuid (EnterpriseCustomer): Enterprise Customer if present then
                send xAPI statements just for this enterprise.
        """
        days = options.get('days', 1)
        enterprise_customer_uuid = options.get('enterprise_customer_uuid')
        enterprise_customer = None

        if enterprise_customer_uuid:
            try:
                enterprise_customer = EnterpriseCustomer.objects.get(uuid=enterprise_customer_uuid)
            except EnterpriseCustomer.DoesNotExist as no_customer_error:
                raise CommandError('Enterprise customer with uuid "{enterprise_customer_uuid}" does not exist.'.format(
                    enterprise_customer_uuid=enterprise_customer_uuid
                )) from no_customer_error

        return days, enterprise_customer

    def handle(self, *args, **options):
        """
        Send xAPI statements.
        """
        if not all((PersistentCourseGrade, CourseOverview)):
            raise NotConnectedToOpenEdX("This package must be installed in an OpenEdX environment.")

        days, enterprise_customer = self.parse_arguments(*args, **options)

        if enterprise_customer:
            try:
                lrs_configuration = XAPILRSConfiguration.objects.get(
                    active=True,
                    enterprise_customer=enterprise_customer
                )
            except XAPILRSConfiguration.DoesNotExist as no_config_exception:
                raise CommandError('No xAPI Configuration found for "{enterprise_customer}"'.format(
                    enterprise_customer=enterprise_customer.name
                )) from no_config_exception

            # Send xAPI analytics data to the configured LRS
            self.send_xapi_statements(lrs_configuration, days)
        else:
            for lrs_configuration in XAPILRSConfiguration.objects.filter(active=True):
                self.send_xapi_statements(lrs_configuration, days)

    def send_xapi_statements(self, lrs_configuration, days):
        """
        Send xAPI analytics data of the enterprise learners to the given LRS.

        Arguments:
            lrs_configuration (XAPILRSConfiguration): Configuration object containing LRS configurations
                of the LRS where to send xAPI  learner analytics.

            days (Numeric):  Deprecated.  Original implementation utilized a "days" parameter to limit
                the number of enrollments transmitted, but this proved to be more problematic than helpful.
        """
        enterprise_course_enrollments = self.get_enterprise_course_enrollments(lrs_configuration.enterprise_customer)
        enterprise_enrollment_ids = self.get_enterprise_enrollment_ids(enterprise_course_enrollments)
        xapi_transmission_queryset = self.get_xapi_transmission_queryset(enterprise_enrollment_ids)
        pertinent_enrollment_ids = self.get_pertinent_enrollment_ids(xapi_transmission_queryset)
        pertinent_enrollments = self.get_pertinent_enrollments(enterprise_course_enrollments, pertinent_enrollment_ids)
        enrollment_grades = self.get_course_completions(pertinent_enrollments)
        users = self.prefetch_users(enrollment_grades)
        course_overviews = self.prefetch_courses(enrollment_grades)
        course_catalog_client = get_course_catalog_api_service_client(site=lrs_configuration.enterprise_customer.site)

        for xapi_transmission in xapi_transmission_queryset:

            object_type = self.get_object_type(xapi_transmission)

            try:
                course_grade = enrollment_grades[xapi_transmission.enterprise_course_enrollment_id]
            except KeyError:
                continue

            user = users.get(course_grade.user_id)
            courserun_id = str(course_grade.course_id)
            course_overview = course_overviews.get(course_grade.course_id)
            course_run_identifiers = course_catalog_client.get_course_run_identifiers(courserun_id)
            course_overview.course_key = course_run_identifiers['course_key']
            course_overview.course_uuid = course_run_identifiers['course_uuid']

            default_error_message = 'Days argument has been deprecated.  Value: {days}'.format(days=days)
            response_fields = {'status': 500, 'error_message': default_error_message}
            response_fields = send_course_completion_statement(
                lrs_configuration,
                user,
                course_overview,
                course_grade,
                object_type,
                response_fields
            )

            if is_success_response(response_fields):
                self.save_xapi_learner_data_transmission_audit(
                    xapi_transmission,
                    course_grade.percent_grade,
                    1,
                    course_grade.passed_timestamp,
                    response_fields.get('status'),
                    response_fields.get('error_message')
                )

    @staticmethod
    def get_object_type(xapi_transmission):
        """
        Returns the object type for the xAPI transmission based on the course object identifier
        """
        object_type = 'course'
        if 'course-v1' in xapi_transmission.course_id:
            object_type = 'courserun'
        return object_type

    def get_enterprise_course_enrollments(self, enterprise_customer):
        """
        Retrieves the set of EnterpriseCourseEnrollment records for the specified EnterpriseCustomer
        """
        return EnterpriseCourseEnrollment.objects.filter(
            enterprise_customer_user_id__enterprise_customer=enterprise_customer
        )

    def get_enterprise_enrollment_ids(self, enterprise_course_enrollments):
        """
        Extracts the identifiers from the provided EnterprsieCourseEnrollment queryset.
        """
        return enterprise_course_enrollments.values_list('id', flat=True)

    @staticmethod
    def get_xapi_transmission_queryset(enterprise_enrollment_ids):
        """
        Retrieves the xAPI transmission records for the specified list of EnterpriseCourseEnrollment identifiers
        """
        return XAPILearnerDataTransmissionAudit.objects.filter(
            enterprise_course_enrollment_id__in=enterprise_enrollment_ids,
            course_completed=0,
        )

    @staticmethod
    def get_pertinent_enrollment_ids(xapi_transmission_queryset):
        """
        Returns the set of enterprise course enrollment identifiers from the
        provided set of xAPI transmission records.
        """
        return xapi_transmission_queryset.values_list('enterprise_course_enrollment_id', flat=True)

    def get_pertinent_enrollments(self, enterprise_course_enrollments, pertinent_enrollment_ids):
        """
        Returns the subset of enterprise course enrollments matching the list of provided identfiers
        """
        return enterprise_course_enrollments.filter(id__in=pertinent_enrollment_ids)

    @staticmethod
    def get_lms_user_id_from_enrollment(enterprise_course_enrollment):
        """
        Retrieves the corresponding User identifier for the specified EnterpriseCourseEnrollment record
        """
        return EnterpriseCustomerUser.objects.get(id=enterprise_course_enrollment.enterprise_customer_user_id).user_id

    @staticmethod
    def get_grade_record_for_enrollment(enterprise_course_enrollment):
        """
        Retrieves the corresponding PersistentCourseGrade record (if available)
        for the specified EnterpriseCourseEnrollment record
        """
        lms_user_id = Command.get_lms_user_id_from_enrollment(enterprise_course_enrollment)
        course_id = enterprise_course_enrollment.course_id
        grade_records = PersistentCourseGrade.objects.filter(
            user_id=lms_user_id,
            course_id=course_id,
            passed_timestamp__isnull=False
        )
        return grade_records.first()

    @staticmethod
    def get_course_completions(enterprise_course_enrollments):
        """
        Get course completions via PersistentCourseGrade for all the learners of given enterprise customer.

        Arguments:
            enterprise_customer (EnterpriseCustomer): Include Course enrollments for learners
                of this enterprise customer.
            days (int): Include course enrollment of this number of days.

        Returns:
            (list): A list of PersistentCourseGrade objects.

        """
        ece_grades = {}
        for ece in enterprise_course_enrollments:
            grade_record = Command.get_grade_record_for_enrollment(ece)
            if grade_record is not None:
                ece_grades.setdefault(ece.id, grade_record)
        return ece_grades

    @staticmethod
    def prefetch_users(enrollment_grades):
        """
        Prefetch Users from the list of user_ids present in the persistent_course_grades.

        Arguments:
            enrollment_grades (list): A list of PersistentCourseGrade.

        Returns:
            (dict): A dictionary containing user_id to user mapping.
        """
        users = User.objects.filter(
            id__in=[grade.user_id for grade in enrollment_grades.values()]
        )
        return {
            user.id: user for user in users
        }

    def prefetch_courses(self, enrollment_grades):
        """
        Prefetch courses from the list of course_ids present in the persistent_course_grades.

        Arguments:
            persistent_course_grades (list): A list of PersistentCourseGrade.

        Returns:
            (dict): A dictionary containing course_id to course_overview mapping.
        """
        return CourseOverview.get_from_ids(
            [grade.course_id for grade in enrollment_grades.values()]
        )

    def save_xapi_learner_data_transmission_audit(self, xapi_transmission,
                                                  course_grade, course_completed, completed_timestamp,
                                                  status, error_message):
        """
        Capture interesting information about the xAPI enrollment (registration) event transmission.

        Arguments:
            xapi_transmission (XAPILearnerDataTransmissionAudit): Transmission audit object being updated
            course_grade (Numeric): Grade value for the enrollment
            course_completed: (Boolean/Numeric): Whether or not the enrollment is considered complete.
            completed_timestamp (Datetime): The point in time when completion occurred.
            status (Numeric):  The response status code
            error_message (String):  Information describing any error state provided by the caller

        Returns:
            None
        """
        xapi_transmission.course_completed = course_completed
        xapi_transmission.completed_timestamp = completed_timestamp
        xapi_transmission.grade = course_grade
        xapi_transmission.status = status
        xapi_transmission.error_message = error_message
        xapi_transmission.save()

        LOGGER.info(
            "[Integrated Channel][xAPI] Successfully updated the XAPILearnerDataTransmissionAudit object with id: {id}"
            ", user: {username} and course: {course_id}".format(
                id=xapi_transmission.id,
                username=xapi_transmission.user.username,
                course_id=xapi_transmission.course_id
            )
        )



"""
Views for the ``pending-enterprise-customer-admin-user`` API endpoint.
"""

from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import filters, permissions

from enterprise import models
from enterprise.api.v1 import serializers
from enterprise.api.v1.permissions import IsInProvisioningAdminGroup
from enterprise.api.v1.views.base_views import EnterpriseReadWriteModelViewSet
from enterprise.logging import getEnterpriseLogger

LOGGER = getEnterpriseLogger(__name__)


class PendingEnterpriseCustomerAdminUserViewSet(EnterpriseReadWriteModelViewSet):
    """
    API views for the ``pending-enterprise-customer-admin-user`` API endpoint.
    Requires staff permissions
    """
    queryset = models.PendingEnterpriseCustomerAdminUser.objects.all()
    filter_backends = (filters.OrderingFilter, DjangoFilterBackend)
    serializer_class = serializers.PendingEnterpriseCustomerAdminUserSerializer
    permission_classes = (permissions.IsAuthenticated, IsInProvisioningAdminGroup)

    FIELDS = (
        'enterprise_customer', 'user_email',
    )
    filterset_fields = FIELDS
    ordering_fields = FIELDS



# Generated by Django 3.2.11 on 2022-02-10 15:46

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields


class Migration(migrations.Migration):

    replaces = [('sap_success_factors', '0001_squashed_0022_auto_20200206_1046'), ('sap_success_factors', '0002_sapsuccessfactorslearnerdatatransmissionaudit_credit_hours'), ('sap_success_factors', '0003_auto_20210701_1556'), ('sap_success_factors', '0004_auto_20210708_1639'), ('sap_success_factors', '0005_sapsuccessfactorsenterprisecustomerconfiguration_prevent_learner_self_submit_grades'), ('sap_success_factors', '0006_sapsuccessfactorsenterprisecustomerconfiguration_idp_id'), ('sap_success_factors', '0007_sapsuccessfactorsenterprisecustomerconfiguration_disable_learner_data_transmissions'), ('sap_success_factors', '0008_alter_sapsuccessfactorsenterprisecustomerconfiguration_enterprise_customer'), ('sap_success_factors', '0009_auto_20220126_1837'), ('sap_success_factors', '0010_sapsuccessfactorsenterprisecustomerconfiguration_display_name'), ('sap_success_factors', '0011_alter_sapsuccessfactorslearnerdatatransmissionaudit_enterprise_course_enrollment_id')]

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('enterprise', '0094_add_use_enterprise_catalog_sample'),
        ('enterprise', '0151_add_is_active_to_invite_key'),
    ]

    operations = [
        migrations.CreateModel(
            name='SAPSuccessFactorsGlobalConfiguration',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('change_date', models.DateTimeField(auto_now_add=True, verbose_name='Change date')),
                ('enabled', models.BooleanField(default=False, verbose_name='Enabled')),
                ('completion_status_api_path', models.CharField(max_length=255)),
                ('course_api_path', models.CharField(max_length=255)),
                ('oauth_api_path', models.CharField(max_length=255)),
                ('search_student_api_path', models.CharField(max_length=255)),
                ('provider_id', models.CharField(default='EDX', max_length=100)),
                ('changed_by', models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL, verbose_name='Changed by')),
            ],
        ),
        migrations.CreateModel(
            name='SAPSuccessFactorsEnterpriseCustomerConfiguration',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('active', models.BooleanField(help_text='Is this configuration active?')),
                ('transmission_chunk_size', models.IntegerField(default=1, help_text='The maximum number of data items to transmit to the integrated channel with each request.')),
                ('channel_worker_username', models.CharField(blank=True, default='', help_text='Enterprise channel worker username to get JWT tokens for authenticating LMS APIs.', max_length=255)),
                ('catalogs_to_transmit', models.TextField(blank=True, default='', help_text='A comma-separated list of catalog UUIDs to transmit. If blank, all customer catalogs will be transmitted. If there are overlapping courses in the customer catalogs, the overlapping course metadata will be selected from the newest catalog.')),
                ('key', models.CharField(blank=True, default='', help_text='OAuth client identifier.', max_length=255, verbose_name='Client ID')),
                ('sapsf_base_url', models.CharField(blank=True, default='', help_text='Base URL of success factors API.', max_length=255, verbose_name='SAP Base URL')),
                ('sapsf_company_id', models.CharField(blank=True, default='', help_text='Success factors company identifier.', max_length=255, verbose_name='SAP Company ID')),
                ('sapsf_user_id', models.CharField(blank=True, default='', help_text='Success factors user identifier.', max_length=255, verbose_name='SAP User ID')),
                ('secret', models.CharField(blank=True, default='', help_text='OAuth client secret.', max_length=255, verbose_name='Client Secret')),
                ('user_type', models.CharField(choices=[('user', 'User'), ('admin', 'Admin')], default='user', help_text='Type of SAP User (admin or user).', max_length=20, verbose_name='SAP User Type')),
                ('additional_locales', models.TextField(blank=True, default='', help_text='A comma-separated list of additional locales.', verbose_name='Additional Locales')),
                ('show_course_price', models.BooleanField(default=False)),
                ('transmit_total_hours', models.BooleanField(default=False, help_text='Include totalHours in the transmitted completion data', verbose_name='Transmit Total Hours')),
                ('enterprise_customer', models.ForeignKey(help_text='Enterprise Customer associated with the configuration.', on_delete=django.db.models.deletion.CASCADE, to='enterprise.enterprisecustomer')),
                ('prevent_self_submit_grades', models.BooleanField(default=False, help_text="When set to True, the integration will use the generic edX service user ('sapsf_user_id') defined in the SAP Customer Configuration for course completion.", verbose_name='Prevent Learner From Self-Submitting Grades')),
                ('idp_id', models.CharField(blank=True, default='', help_text='If provided, will be used as IDP slug to locate remote id for learners', max_length=255)),
                ('disable_learner_data_transmissions', models.BooleanField(default=False, help_text='When set to True, the configured customer will no longer receive learner data transmissions, both scheduled and signal based', verbose_name='Disable Learner Data Transmission')),
                ('display_name', models.CharField(blank=True, default='', help_text='A configuration nickname.', max_length=30)),
            ],
        ),
        migrations.CreateModel(
            name='SapSuccessFactorsLearnerDataTransmissionAudit',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sapsf_user_id', models.CharField(max_length=255)),
                ('enterprise_course_enrollment_id', models.IntegerField(db_index=True)),
                ('course_id', models.CharField(max_length=255)),
                ('course_completed', models.BooleanField(default=True)),
                ('instructor_name', models.CharField(blank=True, max_length=255)),
                ('grade', models.CharField(max_length=100)),
                ('total_hours', models.FloatField(blank=True, null=True)),
                ('completed_timestamp', models.BigIntegerField()),
                ('status', models.CharField(max_length=100)),
                ('error_message', models.TextField(blank=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('credit_hours', models.FloatField(blank=True, null=True)),
            ],
        ),
    ]



# Generated by Django 3.2.12 on 2022-03-25 17:48

from django.db import migrations, models
import django.utils.timezone
import model_utils.fields


class Migration(migrations.Migration):

    dependencies = [
        ('integrated_channel', '0007_delete_learnerdatatransmissionaudit'),
    ]

    operations = [
        migrations.CreateModel(
            name='GenericLearnerDataTransmissionAudit',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('enterprise_customer_uuid', models.UUIDField(blank=True, null=True)),
                ('plugin_configuration_id', models.PositiveIntegerField(blank=True, null=True)),
                ('enterprise_course_enrollment_id', models.PositiveIntegerField(blank=True, db_index=True, null=True)),
                ('course_id', models.CharField(max_length=255)),
                ('course_completed', models.BooleanField(default=True)),
                ('completed_timestamp', models.DateTimeField(blank=True, null=True)),
                ('instructor_name', models.CharField(blank=True, max_length=255)),
                ('grade', models.FloatField(blank=True, null=True)),
                ('total_hours', models.FloatField(blank=True, null=True)),
                ('status', models.CharField(blank=True, max_length=100, null=True)),
                ('error_message', models.TextField(blank=True, null=True)),
            ],
        ),
    ]



"""
Update all courses associated with canvas customer configs to show end dates
"""

from django.apps import apps
from django.contrib import auth
from django.core.management.base import BaseCommand

from integrated_channels.canvas.client import CanvasAPIClient
from integrated_channels.integrated_channel.management.commands import IntegratedChannelCommandMixin

User = auth.get_user_model()


class Command(IntegratedChannelCommandMixin, BaseCommand):
    """
    Update content transmission items to have their respective catalog's uuid.

    ./manage.py lms backfill_course_end_dates
    """
    def handle(self, *args, **options):
        """
        Update all past content transmission items to show end dates.
        """
        options['prevent_disabled_configurations'] = False
        options['channel'] = 'CANVAS'

        ContentMetadataItemTransmission = apps.get_model(
            'integrated_channel',
            'ContentMetadataItemTransmission'
        )

        # get every past transmitted course on canvas channels
        for canvas_channel in self.get_integrated_channels(options):
            transmitted_course_ids = ContentMetadataItemTransmission.objects.filter(
                enterprise_customer=canvas_channel.enterprise_customer,
                integrated_channel_code='CANVAS',
                remote_deleted_at__isnull=True,
            ).values('content_id')

            CanvasAPIClient(canvas_channel).update_participation_types(transmitted_course_ids)



# Generated by Django 3.2.12 on 2022-04-05 23:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('integrated_channel', '0012_alter_contentmetadataitemtransmission_unique_together'),
    ]

    operations = [
        migrations.AlterField(
            model_name='genericlearnerdatatransmissionaudit',
            name='enterprise_course_enrollment_id',
            field=models.IntegerField(blank=True, db_index=True, null=True),
        ),
        migrations.AlterField(
            model_name='genericlearnerdatatransmissionaudit',
            name='plugin_configuration_id',
            field=models.IntegerField(blank=True, null=True),
        ),
    ]



from django.db import migrations
import uuid

def generate_uuid(apps, schema_editor):
    EnterpriseCatalogQuery = apps.get_model('enterprise', 'EnterpriseCatalogQuery')
    for row in EnterpriseCatalogQuery.objects.all():
        row.uuid = uuid.uuid4()
        row.save(update_fields=['uuid'])

class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0127_enterprisecatalogquery_uuid'),
    ]

    operations = [
        migrations.RunPython(generate_uuid, reverse_code=migrations.RunPython.noop),
    ]



"""
Views for the `ChatGPTResponse` API endpoint.
"""
from edx_rbac.decorators import permission_required
from edx_rest_framework_extensions.auth.jwt.authentication import JwtAuthentication
from rest_framework import generics, status
from rest_framework.authentication import SessionAuthentication
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response

from django.shortcuts import get_object_or_404

from enterprise.api.utils import (
    generate_prompt_for_learner_engagement_summary,
    generate_prompt_for_learner_progress_summary,
)
from enterprise.api.v1.serializers import AnalyticsSummarySerializer
from enterprise.models import ChatGPTResponse, EnterpriseCustomer


class AnalyticsSummaryView(generics.GenericAPIView):
    """
    API to generate a signed token for an enterprise admin to use Plotly analytics.
    """
    authentication_classes = [JwtAuthentication, SessionAuthentication]
    permission_classes = (IsAuthenticated,)

    http_method_names = ['post']

    @permission_required(
        'enterprise.can_access_admin_dashboard',
        fn=lambda request, enterprise_uuid: enterprise_uuid
    )
    def post(self, request, enterprise_uuid):
        """
        Generate auth token for plotly.
        """
        role = 'system'
        enterprise_customer = get_object_or_404(EnterpriseCustomer, uuid=enterprise_uuid)

        # Validate payload data
        serializer = AnalyticsSummarySerializer(data=request.data)
        if not serializer.is_valid():
            return Response(
                data={
                    'errors': serializer.errors,
                },
                status=status.HTTP_400_BAD_REQUEST,
            )

        prompt_data = serializer.data
        learner_progress_prompt = generate_prompt_for_learner_progress_summary(prompt_data['learner_progress'])
        learner_engagement_prompt = generate_prompt_for_learner_engagement_summary(prompt_data['learner_engagement'])

        return Response(data={
            'learner_progress': ChatGPTResponse.get_or_create(
                learner_progress_prompt, role, enterprise_customer, ChatGPTResponse.LEARNER_PROGRESS,
            ),
            'learner_engagement': ChatGPTResponse.get_or_create(
                learner_engagement_prompt, role, enterprise_customer, ChatGPTResponse.LEARNER_ENGAGEMENT,
            ),
        })



"""
Django management command to ensure there is at most a single EnterpriseCustomerUser
with `active=True` for each enterprise user.
"""
import logging
from time import sleep

from django.apps import apps
from django.core.management import BaseCommand
from django.db.models import Count

from enterprise.utils import batch

log = logging.getLogger(__name__)


class Command(BaseCommand):
    """
    Django management command to ensure there is at most a single EnterpriseCustomerUser
    with `active=True` for each enterprise user.

    Example usage:
    ./manage.py lms ensure_singular_active_enterprise_customer_user
    """

    def add_arguments(self, parser):
        """
        Entry point for subclassed commands to add custom arguments.
        """
        parser.add_argument(
            '--batch-limit',
            action='store',
            dest='batch_limit',
            default=100,
            help='Number of records in each batch.',
            type=int,
        )

        parser.add_argument(
            '--batch-sleep',
            action='store',
            dest='batch_sleep',
            default=5,
            help='How long to sleep between batches.',
            type=int,
        )

    def _enterprise_customer_user_model(self):
        """ Returns `EnterpriseCustomerUser model """
        EnterpriseCustomerUser = apps.get_model(
            'enterprise',
            'EnterpriseCustomerUser',
        )
        return EnterpriseCustomerUser

    def _historical_enterprise_customer_user_model(self):
        """ Returns `HistoricalEnterpriseCustomerUser model """
        HistoricalEnterpriseCustomerUser = apps.get_model(
            'enterprise',
            'HistoricalEnterpriseCustomerUser',
        )
        return HistoricalEnterpriseCustomerUser

    def _process_user(self, user):
        """
        Finds all EnterpriseCustomerUser objects for the specified user, flipping all but one to
        have `active=False`. The one that is not flipped (remains `active=True`) is based on the
        historical records of which EnterpriseCustomerUser object was flipped to `active=True`
        most recently.
        """
        EnterpriseCustomerUser = self._enterprise_customer_user_model()
        HistoricalEnterpriseCustomerUser = self._historical_enterprise_customer_user_model()
        lms_user_id = user['user_id']
        current_active_ecu_count = user['ecu_count']
        log.info('Processing LMS User ID %s', lms_user_id)
        active_historical_ecus = HistoricalEnterpriseCustomerUser.objects.filter(user_id=lms_user_id, active=True)
        most_recent_active_ecu = active_historical_ecus.order_by('-history_date').first()
        ecus_for_user = EnterpriseCustomerUser.objects.filter(
            user_id=lms_user_id,
            active=True,
        ).exclude(
            id=most_recent_active_ecu.id,
        )
        for ecu in ecus_for_user:
            ecu.active = False
        EnterpriseCustomerUser.objects.bulk_update(ecus_for_user, ['active'])
        log.info(
            'Successfully updated %s of %s EnterpriseCustomerUser objects for LMS User ID %s',
            len(ecus_for_user),
            current_active_ecu_count,
            lms_user_id,
        )

    def handle(self, *args, **options):
        batch_limit = options['batch_limit']
        batch_sleep = options['batch_sleep']

        EnterpriseCustomerUser = self._enterprise_customer_user_model()
        active_ecus = EnterpriseCustomerUser.objects.filter(active=True)
        user_ids_by_active_ecu_count = active_ecus.values('user_id').annotate(ecu_count=Count('user_id'))
        ordered_user_ids_by_active_ecu_count = user_ids_by_active_ecu_count.order_by('-ecu_count')
        user_ids_with_multiple_active_ecus = ordered_user_ids_by_active_ecu_count.filter(ecu_count__gt=1)

        log.info(
            'Found %s enterprise users with mulitple active EnterpriseCustomerUser objects',
            len(user_ids_with_multiple_active_ecus),
        )

        for user_batch in batch(user_ids_with_multiple_active_ecus, batch_size=batch_limit):
            for user in user_batch:
                self._process_user(user)
            sleep(batch_sleep)



# Generated by Django 2.2.24 on 2021-10-13 10:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0144_auto_20211011_1030'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterprisecustomer',
            name='hide_labor_market_data',
            field=models.BooleanField(default=False, help_text='Specifies whether the labor market data should be made known to the learner in learner portal.'),
        ),
        migrations.AddField(
            model_name='historicalenterprisecustomer',
            name='hide_labor_market_data',
            field=models.BooleanField(default=False, help_text='Specifies whether the labor market data should be made known to the learner in learner portal.'),
        ),
    ]



# Generated by Django 4.2.10 on 2024-06-11 03:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0209_alter_enterprisecustomer_disable_expiry_messaging_for_learner_credit_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterprisecustomer',
            name='enable_learner_portal_sidebar_message',
            field=models.BooleanField(default=False, help_text='If checked, learners will be able to see content in the Learner Portal Sidebar found in the HTML box.', verbose_name='Enable learner portal sidebar message'),
        ),
        migrations.AddField(
            model_name='enterprisecustomer',
            name='learner_portal_sidebar_content',
            field=models.TextField(blank=True, help_text='Text shown on the learner portal dashboard for customer specific purposes. Open HTML field.'),
        ),
        migrations.AddField(
            model_name='historicalenterprisecustomer',
            name='enable_learner_portal_sidebar_message',
            field=models.BooleanField(default=False, help_text='If checked, learners will be able to see content in the Learner Portal Sidebar found in the HTML box.', verbose_name='Enable learner portal sidebar message'),
        ),
        migrations.AddField(
            model_name='historicalenterprisecustomer',
            name='learner_portal_sidebar_content',
            field=models.TextField(blank=True, help_text='Text shown on the learner portal dashboard for customer specific purposes. Open HTML field.'),
        ),
        migrations.AlterField(
            model_name='enterprisecustomer',
            name='enable_career_engagement_network_on_learner_portal',
            field=models.BooleanField(default=False, help_text='If checked, the learners will be able to see the link to CEN on the learner portal dashboard.', null=True, verbose_name='Allow navigation to career engagement network from learner portal dashboard'),
        ),
        migrations.AlterField(
            model_name='historicalenterprisecustomer',
            name='enable_career_engagement_network_on_learner_portal',
            field=models.BooleanField(default=False, help_text='If checked, the learners will be able to see the link to CEN on the learner portal dashboard.', null=True, verbose_name='Allow navigation to career engagement network from learner portal dashboard'),
        ),
    ]



# Generated by Django 3.2.23 on 2024-04-08 11:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0204_auto_20240408_0723'),
    ]

    operations = [
        migrations.AlterField(
            model_name='enterprisecustomer',
            name='default_language',
            field=models.CharField(blank=True, choices=[('en', 'English'), ('es-419', 'Español (Latinoamérica)'), ('ar', 'العربية'), ('zh-cn', '中文 (简体)'), ('fr-ca', 'français (Canada)'), ('da', 'dansk'), ('de-de', 'Deutsch (Deutschland)'), ('el', 'Ελληνικά'), ('he', 'עברית'), ('hi', 'हिन्दी'), ('id', 'Bahasa Indonesia'), ('it-it', 'Italiano (Italia)'), ('pt-br', 'Português (Brasil)'), ('pt-pt', 'Português (Portugal)'), ('ru', 'Русский'), ('es-es', 'Español (España)'), ('sw', 'Kiswahili'), ('te', 'తెలుగు'), ('th', 'ไทย'), ('tr-tr', 'Türkçe (Türkiye)'), ('uk', 'українська')], default=None, help_text='Specifies the default language for learners of the organization.', max_length=25, null=True, verbose_name='Learner default language'),
        ),
        migrations.AlterField(
            model_name='historicalenterprisecustomer',
            name='default_language',
            field=models.CharField(blank=True, choices=[('en', 'English'), ('es-419', 'Español (Latinoamérica)'), ('ar', 'العربية'), ('zh-cn', '中文 (简体)'), ('fr-ca', 'français (Canada)'), ('da', 'dansk'), ('de-de', 'Deutsch (Deutschland)'), ('el', 'Ελληνικά'), ('he', 'עברית'), ('hi', 'हिन्दी'), ('id', 'Bahasa Indonesia'), ('it-it', 'Italiano (Italia)'), ('pt-br', 'Português (Brasil)'), ('pt-pt', 'Português (Portugal)'), ('ru', 'Русский'), ('es-es', 'Español (España)'), ('sw', 'Kiswahili'), ('te', 'తెలుగు'), ('th', 'ไทย'), ('tr-tr', 'Türkçe (Türkiye)'), ('uk', 'українська')], default=None, help_text='Specifies the default language for learners of the organization.', max_length=25, null=True, verbose_name='Learner default language'),
        ),
    ]



# Generated by Django 3.2.23 on 2024-04-22 17:09

from django.db import migrations
import fernet_fields.fields


class Migration(migrations.Migration):

    dependencies = [
        ('blackboard', '0019_delete_historicalblackboardenterprisecustomerconfiguration'),
    ]

    operations = [
        migrations.AddField(
            model_name='blackboardenterprisecustomerconfiguration',
            name='decrypted_client_id',
            field=fernet_fields.fields.EncryptedCharField(blank=True, default='', help_text='The API Client ID (encrypted at db level) provided to edX by the enterprise customer to be used to make API calls to Degreed on behalf of the customer.', max_length=255, verbose_name='API Client ID encrypted at db level'),
        ),
        migrations.AddField(
            model_name='blackboardenterprisecustomerconfiguration',
            name='decrypted_client_secret',
            field=fernet_fields.fields.EncryptedCharField(blank=True, default='', help_text='The API Client Secret (encrypted at db level) provided to edX by the enterprise customer to be used to make API calls to Degreed on behalf of the customer.', max_length=255, verbose_name='API Client Secret encrypted at db level'),
        ),
    ]



from django.db import migrations
from datetime import timedelta
    

def _bulk_update_expiration_date(cls, queryset):
    for invite_key in queryset:
        invite_key.expiration_date = invite_key.created + timedelta(days=365)
    cls.objects.bulk_update(queryset, ['expiration_date'])


def backfill_null_expiry_date_for_invite_keys(apps, schema_editor):
    """
    Finds all EnterpriseCustomerInviteKey objects with null expiration_date and applies a default
    expiration_date of 1-year from when the EnterpriseCustomerInviteKey object was created as a
    reasonable fallback such that there are no objects with expiration_date as null so a future
    migration can require this field at the Django admin and database levels.
    """
    EnterpriseCustomerInviteKey = apps.get_model('enterprise', 'EnterpriseCustomerInviteKey')
    HistoricalEnterpriseCustomerInviteKey = apps.get_model('enterprise', 'HistoricalEnterpriseCustomerInviteKey')
    queryset = EnterpriseCustomerInviteKey.objects.filter(expiration_date__isnull=True)
    history_queryset = HistoricalEnterpriseCustomerInviteKey.objects.filter(expiration_date__isnull=True)
    _bulk_update_expiration_date(cls=EnterpriseCustomerInviteKey, queryset=queryset)
    _bulk_update_expiration_date(cls=HistoricalEnterpriseCustomerInviteKey, queryset=history_queryset)

class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0148_auto_20211129_2114'),
    ]

    operations = [
        migrations.RunPython(
            code=backfill_null_expiry_date_for_invite_keys,
            reverse_code=migrations.RunPython.noop,
        ),
    ]




"""
Views for the ``enterprise-user`` API endpoint.
"""

from collections import OrderedDict

from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import permissions, response, status
from rest_framework.pagination import PageNumberPagination

from django.core.exceptions import ValidationError

from enterprise import models
from enterprise.api.v1 import serializers
from enterprise.api.v1.views.base_views import EnterpriseReadOnlyModelViewSet
from enterprise.logging import getEnterpriseLogger

LOGGER = getEnterpriseLogger(__name__)


class EnterpriseCustomerSupportPaginator(PageNumberPagination):
    """Custom paginator for the enterprise customer support."""
    page_size = 8

    def get_paginated_response(self, data):
        """Return a paginated style `Response` object for the given output data."""
        return response.Response(
            OrderedDict([
                ('count', self.page.paginator.count),
                ('num_pages', self.page.paginator.num_pages),
                ('next', self.get_next_link()),
                ('previous', self.get_previous_link()),
                ('results', data),
            ])
        )

    def paginate_queryset(self, queryset, request, view=None):
        """
        Paginate a queryset if required, either returning a page object,
        or `None` if pagination is not configured for this view.

        This method is a modified version of the original `paginate_queryset` method
        from the `PageNumberPagination` class. The original method was modified to
        handle the case where the `queryset` is a `filter` object.
        """
        if isinstance(queryset, filter):
            queryset = list(queryset)

        return super().paginate_queryset(queryset, request, view)


class EnterpriseCustomerSupportViewSet(EnterpriseReadOnlyModelViewSet):
    """
    API views for the ``enterprise-user`` API endpoint.
    """
    queryset = models.EnterpriseCustomerUser.objects.all()
    filter_backends = (DjangoFilterBackend,)
    permission_classes = (permissions.IsAuthenticated,)
    paginator = EnterpriseCustomerSupportPaginator()

    USER_ID_FILTER = 'enterprise_customer_users__user_id'

    def retrieve(self, request, *args, **kwargs):
        """
        - Filter down the queryset of groups available to the requesting uuid.
        """
        enterprise_uuid = kwargs.get('enterprise_uuid', None)
        users = []
        try:
            enterprise_customer_queryset = models.EnterpriseCustomerUser.objects.filter(
                enterprise_customer__uuid=enterprise_uuid,
            )
            users.extend(enterprise_customer_queryset)
            pending_enterprise_customer_queryset = models.PendingEnterpriseCustomerUser.objects.filter(
                enterprise_customer__uuid=enterprise_uuid
            ).order_by('user_email')
            users.extend(pending_enterprise_customer_queryset)
        except ValidationError:
            # did not find UUID match in either EnterpriseCustomerUser or  PendingEnterpriseCustomerUser
            return response.Response(
                {'detail': 'Could not find enterprise uuid {}'.format(enterprise_uuid)},
                status=status.HTTP_404_NOT_FOUND
            )

        users_page = self.paginator.paginate_queryset(
            users,
            request,
            view=self
        )
        serializer = serializers.EnterpriseUserSerializer(
            users_page,
            many=True
        )
        serializer_data = sorted(
            serializer.data, key=lambda k: k['is_admin'], reverse=True)

        return self.paginator.get_paginated_response(serializer_data)



# Generated by Django 4.2.13 on 2024-05-16 16:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0207_alter_enterprisegroupmembership_enterprise_customer_user_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterprisecustomer',
            name='disable_expiry_messaging_for_learner_credit',
            field=models.BooleanField(default=False, help_text='If checked, learners and admins will not receive expiration-related email and other notifications regarding learner credit plans.', verbose_name='Disable expiration messaging for learner credit'),
        ),
        migrations.AddField(
            model_name='historicalenterprisecustomer',
            name='disable_expiry_messaging_for_learner_credit',
            field=models.BooleanField(default=False, help_text='If checked, learners and admins will not receive expiration-related email and other notifications regarding learner credit plans.', verbose_name='Disable expiration messaging for learner credit'),
        ),
    ]



"""
Views for the ``enterprise-course-enrollment`` API endpoint.
"""
from django_filters.rest_framework import DjangoFilterBackend
from edx_rest_framework_extensions.paginators import DefaultPagination
from rest_framework import filters

from django.core.paginator import Paginator
from django.utils.functional import cached_property

from enterprise import models
from enterprise.api.filters import EnterpriseCourseEnrollmentFilterBackend
from enterprise.api.v1 import serializers
from enterprise.api.v1.views.base_views import EnterpriseReadWriteModelViewSet

try:
    from common.djangoapps.util.query import read_replica_or_default
except ImportError:
    def read_replica_or_default():
        return None


class PaginatorWithOptimizedCount(Paginator):
    """
    Django < 4.2 ORM doesn't strip unused annotations from count queries.

    For example, if we execute this query:

        Book.objects.annotate(Count('chapters')).count()

    it will generate SQL like this:

        SELECT COUNT(*) FROM (SELECT COUNT(...), ... FROM ...) subquery

    This isn't optimal on its own, but it's not a big deal. However, this
    becomes problematic when annotations use subqueries, because it's terribly
    inefficient to execute the subquery for every row in the outer query.

    This class overrides the count() method of Django's Paginator to strip
    unused annotations from the query by requesting only the primary key
    instead of all fields.

    This is a temporary workaround until Django is updated to 4.2, which will
    include a fix for this issue.

    See https://code.djangoproject.com/ticket/32169 for more details.

    TODO: remove this class once Django is updated to 4.2 or higher.
    """
    @cached_property
    def count(self):
        return self.object_list.values("pk").count()


class EnterpriseCourseEnrollmentPagination(DefaultPagination):
    django_paginator_class = PaginatorWithOptimizedCount


class EnterpriseCourseEnrollmentViewSet(EnterpriseReadWriteModelViewSet):
    """
    API views for the ``enterprise-course-enrollment`` API endpoint.
    """

    queryset = models.EnterpriseCourseEnrollment.with_additional_fields.all()
    filter_backends = (filters.OrderingFilter, DjangoFilterBackend, EnterpriseCourseEnrollmentFilterBackend)

    USER_ID_FILTER = 'enterprise_customer_user__user_id'
    FIELDS = (
        'enterprise_customer_user', 'course_id'
    )
    filterset_fields = FIELDS
    ordering_fields = FIELDS

    pagination_class = EnterpriseCourseEnrollmentPagination

    def get_queryset(self):
        queryset = super().get_queryset()
        if self.request.method == 'GET':
            queryset = queryset.using(read_replica_or_default())
        return queryset

    def get_serializer_class(self):
        """
        Use a special serializer for any requests that aren't read-only.
        """
        if self.request.method in ('GET',):
            return serializers.EnterpriseCourseEnrollmentWithAdditionalFieldsReadOnlySerializer
        return serializers.EnterpriseCourseEnrollmentWriteSerializer
